/* ReVivAl v0.23 */
/* by Cedric Piette */

/* based on */
/**************************************************************************************************
MiniSat 
http://www.cs.chalmers.se/Cs/Research/FormalMethods/MiniSat/
**************************************************************************************************/


/* *************************** */
/*         OPTIONS             */
/* (un)comment to (un)activate */
/* *************************** */

//#define RANDOMIZED_SEED
//#define PRINT_DETAILS
//#define DONT_LEARN_ANYTHING
//#define PRINT_NUM_ITERATION



/* *************************** */
/*        LIBRARIES            */
/*                             */
/* *************************** */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <zlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include "solver.h"
#include "parsingFunctions.h"

/* *************************** */
/*        CONSTANTS            */
/*   Can be safely modified    */
/* *************************** */

/* Name of the exported file where the redundancy status of each clause is written (1 if redundant, 0 otherwise)
 Used only if EXPORT_REDUNDANCY_STATUS is defined */
#define REDUNDANCY_FILE "/tmp/redundantStatus" 
#define MAX_NUMBER_CLAUSE 100000000
#define MAX_SIZE_LEARNT_CLAUSE 4


/* *************************** */
/*      VARIOUS MACROS         */
/*                             */
/* *************************** */

#define export_Trivial_Unsat_CNF() printf("p cnf 1 2\n1 0\n-1 0\n") 
#define simulate_removeClause(s,c)     s_removeClause(s,c),c=NULL
#define simulate_addItAgain(s,place,c) s_addItAgain(s,c),place=c
#define AmIUndef(lit) (s->assigns[lit_var(lit)]==l_Undef)
#define AmISat(lit) (((s->assigns[lit_var(lit)]==l_False) && (lit&1))  ||  ((s->assigns[lit_var(lit)]==l_True) && (!(lit&1))))
#define negation(lit) (lit^1)
#define free_clause(c) free(c),c=NULL
#define reverseLit(cl,num1,num2) useF1=clause_giveLit(currentClause,num1),clause_setLit(currentClause,num1,clause_giveLit(currentClause,num2)),clause_setLit(currentClause,num2,useF1);


#ifdef PRINT_DETAILS
#define opt_printf(arg) fprintf(stderr,arg);
#define opt_printf_num_current_clause() fprintf(stderr,"clause %d : ",i)
#define opt_printfclause(clause) wellWriteClause_stderr(clause)

#else
#define opt_printf(arg) 
#define opt_printf_num_current_clause()
#define opt_printfclause(clause) 
#endif

#define wellWritten(lit) (lit&1?(-(lit_var(lit)+1)):(lit_var(lit)+1))

int i,globalModif,lastModif,timeOut=0,num_unit=0;

static inline void wellWriteClause(clause *c){
  int i,size=clause_size(c);
  for(i=0;i<size;i++) printf("%d ",wellWritten(clause_giveLit(c,i)));
  printf("\n");
}

static inline void wellWriteClause_stderr(clause *c){
  int i,size=clause_size(c);
  for(i=0;i<size;i++) fprintf(stderr,"%d ",wellWritten(clause_giveLit(c,i)));
}

static inline void s_removeClause(solver *s, clause *c){
  lit* lits = clause_begin(c);
  vecp_remove(solver_read_wlist(s,lit_neg(lits[0])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(lits[1])));
  vecp_remove(solver_read_wlist(s,lit_neg(lits[1])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(lits[0])));
}

static inline void s_addItAgain(solver *s, clause *c){
  vecp_push(solver_read_wlist(s,lit_neg(c->lits[0])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(c->lits[1])));
  vecp_push(solver_read_wlist(s,lit_neg(c->lits[1])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(c->lits[0])));
}

/* After vivification process, enable to return the actual number of (non learnt) clauses of the CNF.
   Before vivification, this function returns s->stats.clauses */
int countLeftClauses(solver *s){
  int i,num=0;
  for(i=0;i<s->stats.clauses;i++)
    if((vecp_begin(&(s->clauses)))[i]) num++;
  return num;
}

int countExportedUnit(solver *s){
  int i,num=0;
  for(i=0;i<s->size;i++)
    if(s->assigns[i]!=l_Undef) num++;
  return num;
}

/* Enables to quickly end vivification process (called if time out is reached) */
void endVivification(){
  fprintf(stderr,"Time out is reached... exporting CNF\n");
  lastModif=globalModif=0;
  timeOut=1;
  return ;
}

/* Update a clause w.r.t. the current assignment 
   (essentially remove falsified literals, and suppress the clause if it contains one satisfied literal, or 
   if it is a unit).
   Return 1 if the clause still exists under the form of a (at least) binary clause, 0 otherwise. */
int refreshClause(solver *s, clause **my_clause){
  int j,k,clause_sat=0;
  for(j=0;j<clause_size(*my_clause);j++){
    if(!AmIUndef(clause_giveLit(*my_clause,j))){
      if(!AmISat(clause_giveLit(*my_clause,j))){
	for(k=(j+1);k<clause_size(*my_clause);k++) 
	  clause_begin(*my_clause)[k-1]=clause_begin(*my_clause)[k];
	(*my_clause)->size_learnt=((clause_size(*my_clause))-1)<<1;
	j-=1;
      }
      else{clause_sat=1; break;}
    }
  }
  if(clause_sat) {
    opt_printfclause(*my_clause);
    opt_printf("  Already SAT\n"); 
    return 0; 
  }
  if(clause_size(*my_clause)==1) {
    num_unit++;
    enqueue(s,clause_giveLit(*my_clause,0),(clause*)0); 
    return 0;
  }
  if(clause_size(*my_clause)==0) {
    fprintf(stderr,"Proved UNSATISFIABLE by ReVivAl !!\n(polynomial instance)\n"); 
    export_Trivial_Unsat_CNF(),exit(0);
  }
  return 1;
}





int vivification(char* CNFfile, int mytime){
  solver* s = solver_new();
  lbool   st;
  FILE *  in;
  veci    learnt_clause;
  clause *currentClause,*learnt_cl=NULL,*c;  
  lit  *begin, *end;
  int j,k,l,lit,noConflict,change=0,useF1,actualCurrentNumClause,
    nbIteration=0,my_lit,num_assuming,p1,p2;
  unsigned int size;
  
#ifdef RANDOMIZED_ALGORITHM
  srand(time(NULL));
#else
  srand(1);
#endif
  
  /* Parse CNF and initialize variables */
  if((in = fopen(CNFfile, "rb"))==NULL) fprintf(stderr, "ERROR! Could not open file: %s\n",CNFfile),exit(1);
  st = parse_DIMACSMin(in, s);
  fclose(in);
  if(MAX_NUMBER_CLAUSE<s->stats.clauses)
    fprintf(stderr,"ReVivAl: too many clauses ! (%d, MAX is %d)\n",(int) s->stats.clauses,MAX_NUMBER_CLAUSE),
      exit(-1);
  if(st==l_False) {
    fprintf(stderr,"Proved UNSATISFIABLE by ReVivAl !!\n(polynomial instance)\n");
    export_Trivial_Unsat_CNF(),exit(0);
  }
  alarm(mytime);
  signal(SIGALRM,endVivification);
  actualCurrentNumClause=lastModif=s->stats.clauses;
  globalModif=1;
  
  /* The flag globalModif is true if literals have been removed from at least one clause.
     Hence, new propagations could be possible, and new subclauses could be produced. 
     All clauses are tested as long as this flag is true (except if a time out is defined and is reached) */
  while(globalModif){
    ++nbIteration;
#ifdef PRINT_NUM_ITERATION
    fprintf(stderr,"Iteration %d\n",nbIteration);
#endif
    globalModif=0;
    for(i=0;i<(globalModif?s->stats.clauses:lastModif);i=((change)?(i):(i+1))){
      opt_printf_num_current_clause();
      if(!(currentClause=(vecp_begin(&(s->clauses)))[i])) { opt_printf("empty\n"); change=0; continue;}
      /* The considered clause is removed from the CNF */
      simulate_removeClause(s,(vecp_begin(&(s->clauses)))[i]);
      change=0,noConflict=1;
      if(solver_propagate(s)) {
	fprintf(stderr,"Proved UNSATISFIABLE by ReVivAl !!\n(polynomial instance)\n"); 
	export_Trivial_Unsat_CNF(),exit(0);
      }
      /* Update the clause w.r.t. the current assignment */
      if(!(refreshClause(s,&currentClause))){
	free_clause(currentClause);
	actualCurrentNumClause--;
	continue;
      }
            
      /* Optional Printing of the original clause */
      opt_printfclause(currentClause);
      size=clause_size(currentClause);
      num_assuming=0;
      /* The MOMS-like sort of the clause has been removed */
      /* MOMS_sort(currentClause); */
      /* Two randomly selected literals are reversed */
      p1=random()%(clause_size(currentClause));
      p2=random()%(clause_size(currentClause));
      reverseLit(currentClause,p1,p2);
      /* Assignment of the opposite of each literal from the clause */
      for(j=0;(j<size) && (noConflict) && (!change);j++){	  
	if(clause_size(currentClause)==1) {break;}
	lit=clause_giveLit(currentClause,j);
	num_assuming++;
	assume(s,negation(lit)); 
	noConflict=((learnt_cl=solver_propagate(s))==NULL);	
	/* The following line can be safely uncommented 
	   (provided that '}' is added just before the "if(!noConflict){" line). 
	   This would remove the next loop in case of conflict. 
	   However, performing this test appears to degrade performances. (weird...) */
	//if(noConflict){
	for(k=j+1;k<size;k++){
	  my_lit=clause_giveLit(currentClause,k);
	  // The literal is assigned to the same value. the remaining part of the clause can be "omitted" 
	  if(s->assigns[lit_var(my_lit)]!=l_Undef){
	    if(AmISat(my_lit)){
	      if(clause_size(currentClause)!=num_assuming+1){
		change=1; 
		clause_begin(currentClause)[num_assuming]=my_lit;
		currentClause->size_learnt=(num_assuming+1)<<1;
	      }else{
		// In this case, the clause is redundant but no shorter clause can be deduced 
		// (last literal already assigned) 
		j=size;
		break;
	      }
	    }
	    // The literal is assigned to the opposite value. it can be removed from the clause 
	    else{ 
	      for(l=k+1;l<size;l++) 
		clause_begin(currentClause)[l-1]=clause_giveLit(currentClause,l); 
	      currentClause->size_learnt=(size-1)<<1; 
	      change=1;
	    }
	    break;
	  }
	} 
      }
      if(!noConflict){
	if(clause_size(currentClause)!=num_assuming){
	  change=1; 
	  currentClause->size_learnt=(num_assuming)<<1;
	}
	veci_new(&learnt_clause);
	veci_resize(&learnt_clause,0);
	solver_analyze(s, learnt_cl, &learnt_clause);
	begin = veci_begin(&learnt_clause);
	end   = begin + veci_size(&learnt_clause);
	if(veci_size(&learnt_clause) > 1){
#ifndef DONT_LEARN_ANYTHING
	  if(veci_size(&learnt_clause) < clause_size(currentClause)){
	    c     = clause_new(s,begin,end,1);
	    vecp_push(&s->learnts,c);
	    s->stats.learnts++;
	  }
#endif
	}
	else{
	  solver_canceluntil(s,s->root_level);
	  num_unit++;
	  enqueue(s,*begin,0);
	}
	if(clause_size(currentClause)!=num_assuming){
	  currentClause->size_learnt=(num_assuming)<<1,change=1;
	}
      }
      solver_canceluntil(s,s->root_level);
      if(clause_size(currentClause)==1){ 
	opt_printf(" --> "); opt_printfclause(currentClause); opt_printf("(unit)\n");
	num_unit++;
	enqueue(s,clause_giveLit(currentClause,0),(clause*)0); 
	free_clause(currentClause);
	actualCurrentNumClause--;
	continue; 
      }
      simulate_addItAgain(s,(vecp_begin(&(s->clauses)))[i],currentClause);
      if(change){
	if(!timeOut) {globalModif=1;  lastModif=i+1;} 
	opt_printf(" --> "); opt_printfclause(currentClause);
      }
      opt_printf("\n");
    }
  }    
  if(solver_propagate(s)) {
    fprintf(stderr,"Proved UNSATISFIABLE by ReVivAl !!\n(polynomial instance)\n"); 
    export_Trivial_Unsat_CNF(),exit(0);
  }
  solver_canceluntil(s,s->root_level);
  
  fprintf(stderr,"#ITERATIONS %d\n",nbIteration);
  fprintf(stderr,"#LEARNT CLAUSES %d\n",(int) s->stats.learnts);
  
  /* Export produced CNF */
  /* Detected unit clauses first, then shortened clauses from the original CNF, then learnt clauses (3 loops) */
  
  printf("p cnf %d %d\n",s->size,(int) ((actualCurrentNumClause)+((int) s->stats.learnts)+countExportedUnit(s)));
  for(i=0,size=s->size;i<size;i++){
    if(s->assigns[i]!=l_Undef){
      printf("%d 0\n",s->assigns[i]==l_True?i+1:-(i+1));
    }      
  }
  for(i=0,size=s->stats.clauses;i<size;i++){
    currentClause=(vecp_begin(&(s->clauses)))[i];
    if(currentClause){
      k=clause_size(currentClause);
      for(j=0;j<k;j++) printf("%d ",wellWritten(clause_giveLit(currentClause,j)));
      printf("0\n");
    }
  }
  for(i=0,size=s->stats.learnts;i<size;i++){
    currentClause=(vecp_begin(&(s->learnts)))[i];
    if(currentClause){
      k=clause_size(currentClause);
      for(j=0;j<k;j++) printf("%d ",wellWritten(clause_giveLit(currentClause,j)));
      printf("0\n");
    }
  }
  return (int) st;
}


int main(int argc, char** argv){
  int mytime=0;
  switch(argc){
  case 3 : mytime=atoi(argv[2]);
  case 2 : break;
  default:  
    fprintf(stderr, "--- ReVivAl v0.23 ---\nusage: %s benchmark.cnf [time]\n",argv[0]),
    fprintf(stderr, "Vivification of the CNF formula 'benchmark.cnf' (DIMACS format)\n"),
    fprintf(stderr, "The resulting CNF is exported in stdout\n\nOption:\n"),
    fprintf(stderr, "time\t Stop the computation after 'time' seconds (no limit if not provided)\n"),
    exit(1);     
  }
  vivification(argv[1],mytime);
  exit(0);
}

