/* ReVivAl v0.23 */
/* by Cedric Piette */

/* based on */
/**************************************************************************************************
MiniSat 
http://www.cs.chalmers.se/Cs/Research/FormalMethods/MiniSat/
**************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <signal.h>
#include <zlib.h>
#include <sys/time.h>
#include <sys/resource.h>


#include "solver.h"
#include "parsingFunctions.h"

//=================================================================================================
// DIMACS Parser:


inline void skipWhitespaceMin(char** in) {
    while ((**in >= 9 && **in <= 13) || **in == 32)
        (*in)++; }

inline void skipLineMin(char** in) {
    for (;;){
        if (**in == 0) return;
        if (**in == '\n') { (*in)++; return; }
        (*in)++; } }

inline int parseIntMin(char** in) {
    int     val = 0;
    int    _neg = 0;
    skipWhitespaceMin(in);
    if      (**in == '-') _neg = 1, (*in)++;
    else if (**in == '+') (*in)++;
    if (**in < '0' || **in > '9') {
      /* Modified to allow "old" DIMACS File (with not '0' at the end of the last clause) */
      fprintf(stderr,"Unexpected char or clause not terminated by a 0 (old DIMACS file)...\n");
      fprintf(stderr,"Considering the file is ended\n");
      return 0;
      /* Original line */
      fprintf(stderr, "PARSE ERROR! Unexpected char: %c\n", **in), exit(1);
    }
    while (**in >= '0' && **in <= '9')
        val = val*10 + (**in - '0'),
        (*in)++;
    return _neg ? -val : val; }

void readClauseMin(char** in, solver* s, veci* lits) {
    int parsed_lit, var;
    veci_resize(lits,0);
    for (;;){
        parsed_lit = parseIntMin(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        veci_push(lits, (parsed_lit > 0 ? toLit(var) : lit_neg(toLit(var))));
		//printf("Litteral : %d -- TRANSFO minisat : %d\n",parsed_lit,(parsed_lit > 0 ? toLit(var) : lit_neg(toLit(var))));
    }
}

lbool parse_DIMACS_mainMin(char* in, solver* s) {
    veci lits;
    veci_new(&lits);

    for (;;){
        skipWhitespaceMin(&in);
        if (*in == 0)
            break;
        else if (*in == 'c' || *in == 'p')
            skipLineMin(&in);
        else{
            lit* begin;
            readClauseMin(&in, s, &lits);
            begin = veci_begin(&lits);
            if (!solver_addclause(s, begin, begin+veci_size(&lits))){
                veci_delete(&lits);
                return l_False;
            }
        }
    }
    veci_delete(&lits);
    return solver_simplify(s);
}

// Reads an input stream to end-of-file and returns the result as a 'char*' terminated by '\0'
// (dynamic allocation in case 'in' is standard input).
//
char* readFileMin(FILE *  in)
{
    char*   data = malloc(65536);
    int     cap  = 65536;
    int     size = 0;

    while (!feof(in)){
        if (size == cap){
            cap *= 2;
            data = realloc(data, cap); }
        size += fread(&data[size], 1, 65536, in);
    }
    data = realloc(data, size+1);
    data[size] = '\0';

    return data;
}


// Inserts problem into solver. Returns FALSE upon immediate conflict.
//
lbool parse_DIMACSMin(FILE * in, solver* s) {
    char* text = readFileMin(in);
    lbool ret  = parse_DIMACS_mainMin(text, s);
    free(text);
    return ret; }


//=================================================================================================

void printStats(stats* stats, int cpu_time){
  return ;
}


